from PIL import Image
import numpy as np
import matplotlib.pyplot as plt

class ThumbnailImage:
    """ Create a thumbnail from an image

    :param tuple extent: Extent of an image (xmin, xmax, ymin, ymax)
    :param data: Image data
    """
    
    extent: list
    im: Image = None
    
    def __init__(self, data, extent, mode='F'):
        self.im = Image.fromarray(data).convert(mode)
        self.extent = extent
        self.xratio = np.abs(self.im.size[0]/(extent[1]-extent[0]))
        self.yratio = np.abs(self.im.size[1]/(extent[3]-extent[2]))
           
    def crop(self, *extent, bgcolor=0):
        """ Change image extent

        :param extent: New extent of an image can be a tuple (xmin, xmax, ymin, ymax), or 4 floats
        :param bgcolor: Color of the thumbnail padding (float or a tuple)
        """
        if len(extent) in [2,5]:
            bgcolor = extent[-1]
        extent = extent[0] if isinstance(extent[0], tuple) else tuple(extent[:4])
        xpix = int(np.round((extent[1]-extent[0])*self.xratio))
        ypix = int(np.round((extent[3]-extent[2])*self.yratio))
        layer = ThumbnailImage(
            extent = extent,
            data = np.full((ypix,xpix),bgcolor).astype(np.asarray(self.im).dtype),
            mode = self.im.mode,
        )
        pos = [int((layer.im.size[i]-self.im.size[i])/2) for i in range(2)]
        layer.im.paste(self.im, pos)
        self.im = layer.im
        self.extent = extent
        self.xratio = np.abs(self.im.size[0]/(extent[1]-extent[0]))
        self.yratio = np.abs(self.im.size[1]/(extent[3]-extent[2]))
        return self

    def resize(self, *resolution):
        """ Change image resolution
        
        :param resolution: Resolution of a new image an be a tuple (xres, yres), or 2 floats
        """
        resolution = resolution[0] if isinstance(resolution[0],tuple) else tuple(resolution)
        self.im = self.im.resize(resolution)
        return self
        
    def draw(self, ax=None, **kwargs):
        """ Draw current image on an axis

        :param ax: Matplotlib axis
        """
        if not ax: ax=plt
        return ax.imshow(np.asarray(self.im), extent=self.extent, origin='lower', **kwargs)



import re
from .list_elements import *

class MaterialBase:
    
    protons: int
    mass: float
    neutrons: int
    electrons: int

    @staticmethod
    def from_string(expr: str):
        # Match elements
        pattern = "([A-Z]+[a-z]?)(\{([0-9]*)([+-]?[0-9]*)\}|)"
        if m := re.match(pattern, expr):
            exceptions = {
                'H': ('H',1,1),
                'D': ('H',1,2),
                'T': ('H',1,3),
            }
            # parse atom and mass numbers
            if not m.group(3) and m.group(1) in exceptions:
                S, Z, A = exceptions[m.group(1)]
            elif m.group(1) in ELEMENTS:
                S = m.group(1)
                Z = ELEMENTS[S][0]
                A = int(m.group(3)) if m.group(3) else Z*2
            else:
                raise Exception("Unknown element:", m.group(1))
            # parse ionization state
            if m.group(4) in ['-','+']:
                I = int(m.group(4)+'1')
            elif m.group(4):
                I = int(m.group(4))
            else:
                I = 0
            # parse mass
            M = ELEMENTS[S][2][str(A)][0] + I * NUCLEONS['[e]'][3]
            return MaterialBase(Z, M, A-Z, Z+I)
        elif expr in NUCLEONS:
            Z, N, E, M, name = NUCLEONS[expr]
            return MaterialBase(Z, M, N, E)
        else:
            raise Exception("Atom canot be parsed from given string:", expr)

    def __init__(self, protons, mass, neutrons=None, electrons=None):
        self.protons   = int(protons)
        self.mass      = float(mass) 
        self.neutrons  = int(protons if neutrons is None else neutrons)
        self.electrons = int(protons if electrons is None else electrons)

    def __str__(self):
        return f"{self.__class__.__name__}(p={self.protons} n={self.neutrons} e={self.electrons} m={self.mass:.3f})"
            
    def __repr__(self):
        return f"{self.__class__.__name__}(p={self.protons} n={self.neutrons} e={self.electrons} m={self.mass:.3f})"
        
    def __add__(self, other):
        protons   = self.protons + other.protons
        mass      = self.mass + other.mass
        neutrons  = self.neutrons + other.neutrons
        electrons = self.electrons + other.electrons
        return MaterialBase(protons, mass, neutrons, electrons)

    def __mul__(self, other):
        protons   = self.protons * other
        mass      = self.mass * other
        neutrons  = self.neutrons * other
        electrons = self.electrons * other
        return MaterialBase(protons, mass, neutrons, electrons)
        



from reportlab.platypus import Table, Paragraph, Spacer, PageBreak
from reportlab.lib.units import inch
import numpy as np

from ..settings import *
from ....settings import ROOT_SOURCE
from ....environment import Environment

class UnitsSection:
    
    env: Environment
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, env: Environment):
        self.env = env
        
    def parse_table(self):
        TABLE_STYLE = [
            ('GRID',       (0,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('BACKGROUND', (0,0), (-1,0),   PALETTE['node_name']),  
            ('BACKGROUND', (0,1), (0,-1),   PALETTE['prop_name']),  
            ('BACKGROUND', (1,1), (-1,-1),    PALETTE['prop_value']),  
        ]
    
        data = [   
            ['Name','Value','Units','Source'],
        ]
        for name, unit in self.env.units.items():
            # create source link
            if ROOT_SOURCE in unit['source']:
                link_source = f"#source_{unit['source']}" 
            else:
                link_source = f"#source_{unit['source']}_{unit['lineno']}" 
            src = Paragraph(f"<a href=\"{link_source}\" color=\"blue\">{unit['source']}:{unit['lineno']}</a>")
            # add row to the table
            data.append([name, unit['value'], unit['units'], src])
            
        colWidths = list(np.array([0.2,0.2,0.2,0.4])*(PAGE_WIDTH-2*inch))
        return Table(data, style=TABLE_STYLE, hAlign='LEFT', colWidths=colWidths)
        
    def parse(self):
        blocks = []
        blocks.append(PageBreak())
        blocks.append(Paragraph(f"Custom units", SECTION_STYLE))
        blocks.append(self.parse_table()) 
        blocks.append(Spacer(1,0.2*inch))
        return blocks
    


from reportlab.platypus import Paragraph, Table, Spacer, PageBreak
from reportlab.lib.units import inch
import numpy as np
import re

from ..settings import *
from ....nodes import Node, BooleanNode, IntegerNode, FloatNode, StringNode, ModNode
from ....settings import Order, Sign, Keyword, EnvType, DocsType
from ....environment import Environment

class NodeSection:
    
    names: list
    nodes: list
    env: Environment
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, names, nodes, env):
        self.names = names
        self.nodes = nodes
        self.env = env
        pass

    def _init_bool(self, node):
        self.dtype = node.keyword
        if node.value:
            self.value = Keyword.TRUE if node.value.value else Keyword.FALSE
            
    def _init_int(self, node):
        if node.units_raw:
            self.options = [f"{option.value.value} {option.value.unit}" for option in node.options]
        else:
            self.options = [f"{option.value.value}" for option in node.options]
        self.dtype = node.keyword
        if node.value:
            if isinstance(node.value.value, (np.ndarray, list)):
                self.value = str(node.value.value)
            else:
                self.value = str(int(node.value.value))
        if node.unsigned:
            self.dtype = f"u{self.dtype}"
        if node.precision:
            self.dtype = f"{self.dtype}{node.precision}"
            
    def _init_float(self, node):
        if node.units_raw:
            self.options = [f"{option.value.value} {option.value.unit}" for option in node.options]
        else:
            self.options = [f"{option.value.value}" for option in node.options]
        self.dtype = node.keyword
        if node.value:
            if isinstance(node.value.value, (np.ndarray, list)):
                self.value = str(node.value.value)
            else:
                exp = np.log10(node.value.value)
                if exp<=3 or exp>=-3:
                    self.value = f"{node.value.value:.03f}"
                else:
                    self.value = f"{node.value.value:.03e}"
        if node.precision:
            self.dtype = f"{self.dtype}{node.precision}"
            
    def _init_str(self, node):
        self.options = [str(option.value.value) for option in node.options]
        self.dtype = node.keyword
        if node.value:
            self.value = str(node.value.value)
            
    def _init_mod(self, node):
        self.value = node.value
        self.dtype = node.keyword
            
    def _init_node(self, node):
        # find out units
        self.unit = ''
        if node.keyword != ModNode.keyword and node.value and node.value.unit:
            self.unit = node.value.unit
        elif node.units_raw:
            self.unit = node.units_raw
            
        """
        parent_name = ".".join(node.name.split(".")[:-1])
        if m := re.match(f".*({Sign.CONDITION}[0-9]+)$", parent_name):
            cid = m.group(1)
            case = env.branching.cases[m.group(1)]
            if case.expr is None:
                self.condition = None
            else:
                condition = case.expr.replace("{","<font color='orange'>{")
                condition = condition.replace("}","}</font>")
                self.condition = condition
        else:
            self.condition = None
        """
        # condition
        self.condition = None
        if node.condition:
            condition = node.condition.replace("{","<font color='orange'>{")
            condition = condition.replace("}","}</font>")
            self.condition = condition
        
        # type specific initialization
        self.value = None
        self.options = None
        getattr(self,f"_init_{node.keyword}")(node)
        if self.options:
            self.options = ", ".join(self.options)
        
        # additional settings
        if node.constant:
            self.dtype = f"constant {self.dtype}"
        self.dformat = node.format if node.keyword == StringNode.keyword else None
        if node.keyword == ModNode.keyword or node.tags is None:
            self.tags = ''
        else:
            self.tags = ", ".join(node.tags)
        self.description = None if node.keyword ==ModNode.keyword else node.description
            
    def parse_node(self, name, node):
    
        self._init_node(node)
    
        if DocsType.DEFINITION|DocsType.MODIFICATION in node.docs_type:
            self.color = PALETTE['def/mod']
        elif DocsType.DEFINITION in node.docs_type:
            self.color = PALETTE['def']
        elif DocsType.DECLARATION|DocsType.MODIFICATION in node.docs_type:
            self.color = PALETTE['dec/mod'] 
        elif DocsType.DECLARATION in node.docs_type:
            self.color = PALETTE['dec']
        elif DocsType.MODIFICATION in node.docs_type:
            self.color = PALETTE['mod']
    
        # define table style
        TABLE_STYLE = [
            # whole grid
            #('GRID',       (0,0), (-1,-1),  0.5,     colors.goldenrod), 
            ('GRID',       (1,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('FONTNAME',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontName),
            ('FONTSIZE',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontSize),
            # top panel
            ('FONTSIZE',   (0,0), (-1,-1),  TABLE_HEADER_STYLE.fontSize),
            ('TEXTCOLOR',  (0,0), (-1,0),   PALETTE['c4']),   
            ('BACKGROUND', (0,0), (-1,0),   PALETTE['node_name']),
            ('ALIGN',      (-2,0),(-1,0),   'RIGHT'),
            ('BACKGROUND', (1,1), (1,-1),   PALETTE['prop_name']),  
            ('BACKGROUND', (2,1), (-1,-1),  PALETTE['prop_value']),   
            ('SPAN',       (1,0), (2,0)     ),                            # parameter name
            ('BACKGROUND', (0,0), (0,-1),   self.color),                  # instance type
        ]
        
        # construct a node table
        pname = Paragraph(f"<strong>{name}</strong>", )
        source = Paragraph(f"<a href=\"#source_{node.source}_{node.lineno}\" color=\"blue\">{node.source}:{node.lineno}</a>")
        data = [
            ['', source, '', self.dtype],
        ]
        if self.value:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Value:', Paragraph(self.value, TABLE_BODY_STYLE)])
        if self.unit:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Default Unit:',  Paragraph(self.unit, TABLE_BODY_STYLE)])
        if self.condition:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Condition:',     Paragraph(self.condition, style=TABLE_BODY_STYLE)])
        if self.tags:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Tags:',          Paragraph(self.tags, TABLE_BODY_STYLE)])
        if self.options: 
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Options:',       Paragraph(self.options, TABLE_BODY_STYLE)])
        if self.dformat:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Format:',        Paragraph(self.dformat, TABLE_BODY_STYLE)])
        if self.description:
            TABLE_STYLE.append(('SPAN', (2,len(data)), (-1,len(data)) ))
            data.append(['','Description:',   Paragraph(self.description, TABLE_BODY_STYLE)])
            
        colWidths = list(np.array([0.01,0.2, 0.57, 0.22])*(PAGE_WIDTH-2*inch))
        return Table(data,style=TABLE_STYLE, hAlign='LEFT', colWidths=colWidths)

    def parse(self):
        blocks = []
        blocks.append(PageBreak())
        blocks.append(Paragraph(f"Node list", SECTION_STYLE) )
        for name in self.names:
            blocks.append(Spacer(1,0.1*inch))
            blocks.append(Paragraph(f"<strong>{name}</strong><a name=\"node_{name}\"></a>"))
            blocks.append(Spacer(1,0.1*inch))
            for node in self.nodes[name]:
                blocks.append(self.parse_node(name, node))
        blocks.append(Spacer(1,0.2*inch))
        return blocks


from reportlab.platypus import Table, Paragraph, Spacer
from reportlab.lib.units import inch
import numpy as np

from ..settings import *

class TypesSection:
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def parse_table(self):
        TABLE_STYLE = [
            ('BACKGROUND', (0,0), (0,0),   PALETTE['dec']),  
            ('BACKGROUND', (0,1), (0,1),   PALETTE['def']),  
            ('BACKGROUND', (0,2), (0,2),   PALETTE['dec/mod']),  
            ('BACKGROUND', (0,3), (0,3),   PALETTE['def/mod']),  
            ('BACKGROUND', (0,4), (0,4),   PALETTE['mod']),  
        ]
    
        data = [   
            ['','Declaration'],
            ['','Definition'],
            ['','Declaration / Modification'],
            ['','Definition / Modification'],
            ['','Modification'],
        ]
        colWidths = list(np.array([0.05,0.95])*(PAGE_WIDTH-2*inch))
        return Table(data, style=TABLE_STYLE, hAlign='LEFT', colWidths=colWidths)
        
    def parse(self):
        blocks = []
        blocks.append(Paragraph(f"Node types", SECTION_STYLE))
        blocks.append(self.parse_table()) 
        blocks.append(Spacer(1,0.2*inch))
        return blocks
    


from reportlab.platypus import Paragraph, Table, Spacer
from reportlab.lib.units import inch
import numpy as np

from ..settings import *
from ....settings import DocsType

class ReferenceSection:
    
    names: list
    nodes: list
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
        
    def __init__(self, names, nodes):
        self.names = names
        self.nodes = nodes
    
    def parse_table(self):
        TABLE_STYLE = [
            ('GRID',       (0,1), (-1,-1),  0.5, PALETTE['prop_name']),
            ('BACKGROUND', (0,0), (0,0),   PALETTE['node_name']),  
            ('BACKGROUND', (0,1), (-1,-1),   PALETTE['prop_value']),  
            ('BACKGROUND', (1,0), (1,0),   PALETTE['dec']),  
            ('BACKGROUND', (2,0), (2,0),   PALETTE['def']),  
            ('BACKGROUND', (3,0), (3,0),   PALETTE['dec/mod']),  
            ('BACKGROUND', (4,0), (4,0),   PALETTE['def/mod']),  
            ('BACKGROUND', (5,0), (5,0),   PALETTE['mod']),  
        ]
    
        data = [['Property name', '#', '#', '#', '#', '#']]
        for name in self.names:
            p = Paragraph(f"<a href=\"#node_{name}\" color=\"blue\">{name}</a>")
            counts = [0, 0, 0, 0, 0]
            for node in self.nodes[name]:
                if DocsType.DEFINITION|DocsType.MODIFICATION in node.docs_type:
                    counts[3] += 1
                elif DocsType.DEFINITION in node.docs_type:
                    counts[1] += 1
                elif DocsType.DECLARATION|DocsType.MODIFICATION in node.docs_type:
                    counts[2] += 1
                elif DocsType.DECLARATION in node.docs_type:
                    counts[0] += 1
                elif DocsType.MODIFICATION in node.docs_type:
                    counts[4] += 1
            counts = ['' if c==0 else c for c in counts]
            data.append([p]+counts)
        colWidths = list(np.array([0.8, 0.04, 0.04, 0.04, 0.04, 0.04])*(PAGE_WIDTH-2*inch))
        return Table(data, style=TABLE_STYLE, hAlign='LEFT', colWidths=colWidths)
        
    def parse(self):
        blocks = []       
        blocks.append(Paragraph(f"Node reference", SECTION_STYLE))
        blocks.append(self.parse_table())
        blocks.append(Spacer(1,0.2*inch))
        return blocks


from reportlab.lib import colors
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.rl_config import defaultPageSize

SAMPLE_STYLE_SHEET = getSampleStyleSheet()

FONT_NAME = 'Helvetica'
FONT_SIZE = 10
PAGE_HEIGHT=defaultPageSize[1]; 
PAGE_WIDTH=defaultPageSize[0]
        
SECTION_STYLE = ParagraphStyle(
    "SectionTitleStyle",
    parent=SAMPLE_STYLE_SHEET['Normal'],
    fontName = FONT_NAME,
    fontSize = 14,
    spaceAfter = 10
)

GROUP_STYLE = ParagraphStyle(
    "CustomStyle",
    parent=SAMPLE_STYLE_SHEET["Normal"],
    fontName=FONT_NAME,
    fontSize=12,
    spaceBefore=0, 
    spaceAfter=10,  # Add 20 points of space after the paragraph
)

TABLE_HEADER_STYLE = ParagraphStyle(
    "CustomStyle",
    parent=SAMPLE_STYLE_SHEET["Normal"],
    fontName=FONT_NAME,
    fontSize=FONT_SIZE,
    textColor=colors.saddlebrown
)

TABLE_BODY_STYLE = ParagraphStyle(
    "CustomStyle",
    parent=SAMPLE_STYLE_SHEET["Normal"],
    fontName=FONT_NAME,
    fontSize=FONT_SIZE,
)

CASE_STYLE = [
    # whole grid
    #('GRID',       (0,0), (-1,-1),  0.5,     colors.goldenrod),
    ('FONTNAME',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontName),
    ('FONTSIZE',   (0,0), (-1,-1),  TABLE_BODY_STYLE.fontSize),
    # top panel
    ('FONTSIZE',   (0,0), (-1,-1),  TABLE_HEADER_STYLE.fontSize),
    #('BACKGROUND', (0,0), (1,0),    colors.lightgreen),    
    #('BACKGROUND', (1,0), (-1,0),   colors.floralwhite),  
]

PALETTE = {
    'dec':     '#698D3F',
    'def':     '#AABA78',
    'dec/mod': '#D77F33',
    'def/mod': '#DDA15E',
    'mod':     '#FBE974',
    'c4': colors.saddlebrown, #'#',
    'node_name': '#E7CCB1', #colors.navajowhite, 
    'prop_name': '#FAEDCD', #colors.antiquewhite, #'#',
    'prop_value': '#FEFAE0', #colors.floralwhite, #'#',
}


from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Table, KeepTogether 
from reportlab.lib.units import inch
from reportlab.platypus.flowables import KeepTogether 
import numpy as np

from .settings import *
from .sections.node import NodeSection
from .sections.types import TypesSection
from .sections.reference import ReferenceSection
from .sections.units import UnitsSection
from .sections.sources import SourcesSection
from ...environment import Environment
from ...settings import Order, Sign, Keyword, EnvType, DocsType
from ...lists import NodeList
from ...nodes import Node, BooleanNode, IntegerNode, FloatNode, StringNode, ModNode

class ExportPDF:
    
    env: Environment
    nodes: dict         # list of nodes grouped by name
    names: list         # sorted list of node names
    
    def __enter__(self):
        return self
    
    def __exit__(self, type, value, traceback):
        pass
    
    def __init__(self, env: Environment, **kwargs):
        if env.envtype != EnvType.DOCS:
            raise Exception("Given environment is not a documentation environment")
        self.env = env
        
        # group nodes according to their names
        nodes = self.env.nodes.query("*")
        self.nodes = {}
        for node in nodes:
            name = node.clean_name()
            if name in self.nodes:
                self.nodes[name].append(node)
            else:
                self.nodes[name] = [node]

        # create a sorted list of node names
        self.names = list(self.nodes.keys())
        self.names.sort()
        
    def build(self, file_path: str, title, pageinfo):
        
        def myFirstPage(canvas, doc):
            canvas.saveState()
            canvas.setFont('Times-Bold',16)
            canvas.drawCentredString(PAGE_WIDTH/2.0, PAGE_HEIGHT-108, title)
            canvas.setFont('Times-Roman',9)
            canvas.drawString(inch, 0.75 * inch, "First Page / %s" % pageinfo)
            canvas.restoreState()
            
        def myLaterPages(canvas, doc):
            canvas.saveState()
            canvas.setFont('Times-Roman',9)
            canvas.drawString(inch, 0.75 * inch, "Page %d %s" % (doc.page, pageinfo))
            canvas.restoreState()

        doc = SimpleDocTemplate(file_path)
        blocks = [Spacer(1,1*inch)]

        # list node types
        with TypesSection() as tmpl:
            blocks += tmpl.parse()    

        # create a quick reference of nodes with links
        with ReferenceSection(self.names, self.nodes) as tmpl:
            blocks += tmpl.parse()
        
        # list all nodes and their properties
        with NodeSection(self.names, self.nodes, self.env) as tmpl:
            blocks += tmpl.parse()
 
        # list units
        with UnitsSection(self.env) as tmpl:
            blocks += tmpl.parse()    
 
        # list sources
        with SourcesSection(self.env) as tmpl:
            blocks += tmpl.parse()    
 
        # build a documentation
        doc.build(blocks, onFirstPage=myFirstPage, onLaterPages=myLaterPages)    
